"""
This file contains all of the necessary functions to perform EMW and PWM for
monotone allocations, as described in Appendix C.

NOTE: Gurobi is required to run these functions.
"""

import numpy as np
from gurobipy import *
import matplotlib.pyplot as plt

            
def EWM(data, prop, level = 1, rule = 1, message = 0, dm = 0):
    """Performs EWM over monotone allocations, as described in the appendix.
    
    Args:
        data: a numpy array corresponding to the dataset to be used, in 
            (Y, D, X) format
        prop: the propensity score P(D = 1)
        level: specifies the class G_k
        rule: a boolean flag which separates the "zero threshold" case 
            (treat everyone/no one) from the "positive threshold" case.
        message: boolean flag which specifies if the gurobi output should be
            printed
        dm: a boolean flag which specifies if the outcome should be demeaned as
            in Remark 3.2
    Returns:
        A vector which correponds to tau_i for each individual in the sample
        A vector which corresponds to the assignment of each individual in the sample
        A numpy array which corresponds to the optimal rule   
    """
    n = len(data)
    x = data[:,2:]
    dim = len(x[0])
    y = data[:,0]
    ym = np.mean(y)
    #this is the de-meaned version of the outcome
    ydm = y - ym
    d = data[:,1]
    
    if dm == 0:
        g = ((1-prop)*y*d - prop*y*(1-d))
        #g = (y*d/prop - y*(1-d)/(1-prop))
    else:
        g = ((1-prop)*ydm*d - prop*ydm*(1-d))
            
    #Big-M value
    C_gen = (2**(level - 1) + 1)*21
    
    if message == 1:
        print("setup...")
        
    if rule == 0: 
        obj = np.sum(g)
        if obj > 0:
            if message == 1:
                print("all")
                print("objective:",np.dot(g,np.ones(n)))
            return g, np.ones(n),np.array([np.concatenate(([1],np.zeros(dim)))])
        else:
            if message == 1:
                print("nothing")
                print("objective:",np.dot(g,np.zeros(n)))
            return g, np.zeros(n),np.array([np.concatenate(([-1],np.zeros(dim)))])
        
    elif rule == 1:
        # create the LP object, set up as a maximization problem
        m = Model("EWM")
        
        if message != 1:
            m.setParam('OutputFlag', False)
            
        index_vector = range(0, n, 1)
        T = 2**(level - 1)
        M = 2**(level - 1) + 1 #index starts at 0 
        beta_dim = range(0, M + 1, 1) #add extra component for constant term
        
        # set up decision variables
        z = {}
        for i in index_vector:
            z[i] = m.addVar(vtype = GRB.BINARY, name="z%s"%i)
                    
        beta = {}
        for i in beta_dim:
            beta[i] = m.addVar(ub = 20,lb = -20, name="beta%s"%i)
            
        m.update()
        
        # set up objective
        m.setObjective(quicksum(g[iv]*z[iv] for iv in index_vector),GRB.MAXIMIZE)
        
        # add constraints:
        D_m = make_DM(T)
        B = list(beta.values())
        cov = np.zeros(M + 1)
        for iv in index_vector:
            for i in range(0, M + 1, 1):
                if i == 0:
                    cov[i]= x[iv][1]
                else:
                    cov[i] = f(T, i - 1, x[iv][0], 15, 5)
            C = C_gen
            m.addConstr(np.dot(cov, B) - z[iv]*C + C*10**(-7) <= 0,"constraint A%d"%iv)
            m.addConstr(C*z[iv] - (C*(1) + np.dot(cov,B)) <= 0,"constraint B%d"%iv)
        m.addConstr(beta[0] == - 1)
        mon_constr = np.dot(D_m, B[1:])
        for i in range(0, len(mon_constr), 1):
            m.addConstr(mon_constr[i] <= 0)
        
        if message == 1:
            print("solving...")
       
        #set Gurobi tuning parameters:
        m.setParam('IntFeasTol',1e-9)
        m.setParam('FeasibilityTol',1e-9)        
        m.setParam('Cuts',3)
        m.setParam('VarBranch',0)
        m.setParam('Heuristics',0.001)	
        
        #If you need to save nodes to external file:
        #m.setParam('NodefileStart',0.5)
        #m.setParam('NodefileDir',"/sscc/home/m/mtu579/nodes/file%d"%level)
        
        m.optimize()
        
        z_vec = [z[iv].x for iv in index_vector]
        b_vec = [beta[i].x for i in beta_dim]
        return g, z_vec, b_vec
        
def PWM(data, m, prop, level = 0, message = 0, deme=0):
    """Computes a candidate PWM rule and its corresponding complexity penalty
    AT A SPECIFIED K. Does not compute the complexity-optimal rule over all k.
    
    Args:
        data: data: a numpy array corresponding to the dataset to be used, in 
            (Y, D, X) format
        m: The size of the test dataset
        prop: the propensity score P(D = 1)
        level: specifies the class G_k
        message: boolean flag which specifies if the gurobi output should be
            printed
        deme: a boolean flag which specifies if the outcome should be demeaned as
            in Remark 3.2
    """
    def _pen(data, rule, prop, ltd):
        """computes the empirical welfare on the holdout sample"""
        y = data[:,0]
        ym = np.mean(y)
        #this is the de-meaned version of the outcome
        ydm = y - ym
        d = data[:,1]
        if deme == 0:
            g = ((1-prop)*y*d - prop*y*(1-d))
        else:
            g = ((1-prop)*ydm*d - prop*ydm*(1-d))
        z = np.zeros(len(data))
        for i in range(0,len(data),1):
            x = data[i,2:]
            z[i] = _is_in(rule,x)
        return np.dot(g,z)
    
    n = len(data)
    len_train_data = n-m
    train_data = data[0:len_train_data]
    pen_data = data[len_train_data:n]
    [g,z,b] = EWM(train_data, prop, level, 1, message, dm=deme)    
    c = _pen(pen_data, b, prop, len_train_data)
    return c, z, b
    
def f(T, j, x, c = 1, d = 0):
    """Implements the psi function as defined in the appendix"""
    if T*(x-d) >= (j-1)*c and T*(x-d) <= (j+1)*c and x >= 0+d and x <= c+d:
        return 1 - np.abs((T/c)*(x-d) - j)
    else:
        return 0
        
def make_DM(T):
    """Constructs the differentiation matrix as defined in Example 3.1"""
    D_m = np.zeros((T, T + 1))
    for i in range(0, T, 1):
        for j in range(0, T + 1, 1):
            if i==j:
                D_m[i,j] = -1
            elif i == j - 1:
                D_m[i,j] = 1
    return D_m

def plot_rule(data, z, b_vec):
    """Plots a given monotone rule over a scatterplot of the covariates.
    
    Args: 
        data: a numpy array corresponding to the dataset to be used, in 
            (Y, D, X) format
        z: a vector of binary variables which specify if the observations are 
        treated/not treated
        b_vec: the given rule
    
    Returns:
        None (plot displayed)
    """
    covs = data[:,2:]
    x = np.arange(0,20,.1)
    n = len(x)
    y = np.zeros(n)
    
    for i in range(0,n,1):
        for j in range(1,len(b_vec),1):
            y[i] = y[i] + b_vec[j]*f(len(b_vec)-2,j-1,x[i],15,5)

    plt.title('EWM Allocation')
    plt.xlabel('Years of Education')   
    plt.ylabel('Pre-program Annual Earnings')
    plt.xlim(5,19)
    plt.ylim(0,25000)
    
    use_colors = {0:"tomato",1:"darkgreen"}
    #plt.scatter(covs[:,0],covs[:,1],c = [use_colors[p] for p in z])
    plt.scatter(covs[:,0], covs[:,1])
    
    plt.plot(x,y)
        

def _is_in(rule, x):
    """determines if a given covariate point is treated given rule"""
    cov = np.zeros(len(rule))
    for i in range(0,len(rule), 1):
        if i == 0:
            cov[i]= x[1]
        else:
            #use M-1 here because of the way f is defined
            cov[i] = f(len(rule) - 2, i - 1, x[0], 15, 5)
    if np.dot(cov,rule) >= 0:
        return 1
    else:
        return 0
        
        
        
        
        
        
        
