#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This file re-creates Table 1 in Appendix B.1.

NOTE: Gurobi is required to run this file.
"""

import PWM_app as PWM
import numpy as np
import pandas as pd
from scipy.stats import norm

#set seed
np.random.seed(5)

#propensity score from JTPA
prop = 2/3

#read in file
data = pd.io.stata.read_stata("/Users/max/Dropbox/github/SWM/jtpa2.dta")

#grab relevant columns
y = data['earnings'].values
d = data['assignmt'].values
x1 = data['prevearn'].values
x2 = data['bfeduca'].values
x2 = x2.astype(np.int)
#scale data
ys = y/10000
x1s = x1/10000
x2s = x2
#scaled and unscaled datasets
data = np.dstack((y,d,x2,x1))[0]
data_scaled = np.dstack((ys,d,x2s,x1s))[0]


#permute indices just incase sorted
p = np.random.permutation(len(data))
data = data[p]
data_scaled = data_scaled[p]

#split estimation and test data
data_est = data_scaled[0:5505]
data_test = data[5505:]
data_first = data[0:5505]


"""FUNCTIONS TO DO INFERENCE IN TEST DATASET"""
def mean_welfare(rule, data, is_in = 0):
    """Computes the estimated welfare as described in Appendix B.2.
    
    Args: 
        rule: the decision rule at which to evaluate welfare
        data: the data used to evaluate welfare
        in_in: an optional parameter which can specify the full vector of 
            assignments (if these have been pre-computed).
    
    Returns:
        An estimate of the welfare
    """
    x = data[:,2:]
    y = data[:,0]
    d = data[:,1]
    #g = (y*d/prop - y*(1-d)/(1 - prop))
    g_1 = y*d/prop
    g_2 = y*(1-d)/(1-prop)
    if isinstance(is_in, int) == 1:
        is_in = np.array([PWM._is_in(rule, i) for i in x])
    return np.mean(g_1*is_in + g_2*(1 - is_in))

def var_est(rule, data, is_in = 0):
    """Computes an estimate of the asymptotic variance for 
    the estimated welfare as described in Remark B.1.
    
    Args: 
        rule: the decision rule at which to evaluate welfare
        data: the data used to evaluate welfare
        in_in: an optional parameter which can specify the full vector of 
            assignments (if these have been pre-computed).
    
    Returns:
        An estimate of the asymptotic variance
    """
    x = data[:,2:]
    y = data[:,0]
    d = data[:,1]
    g_1 = y*d/prop
    g_2 = y*(1-d)/(1-prop)
    if isinstance(is_in, int) == 1:
        is_in = np.array([PWM._is_in(rule, i) for i in x])
    return (np.mean((g_1*is_in + g_2*(1 - is_in))**2) - np.mean(g_1*is_in + g_2*(1 - is_in))**2)

def welfare_difference(rule_1, rule_2, null, data):
    """Computes a tests statistic for the null hypothesis 
    that two decision rules deliver the same welfare.
    
    Args: 
        rule_1: the first decision rule to evaluate
        rule_2: the second decision rule to evaluate
        null: the difference to test against (typically zero)
        data: the data used to run the test
    
    Returns:
        The t-statistic for the difference in means
    """
    n = len(data)
    mean_rule_1 = mean_welfare(rule_1, data)
    mean_rule_2 = mean_welfare(rule_2, data)
    var_rule_1 = var_est(rule_1, data)
    var_rule_2 = var_est(rule_2, data)
    cov = cov_est(rule_1, rule_2, data)
    num = np.sqrt(n)*(mean_rule_1 - mean_rule_2 - null)
    denom = np.sqrt(var_rule_1 - 2*cov + var_rule_2)
    return num/denom

def cov_est(rule_1, rule_2, data):
    """Computes an estimate of the asymptotic covariance for the joint distribution
    of the welfare for two decision rules (this is used to compute the standard
    errors for the random baseline below).
    
    Args: 
        rule_1: the first decision rule to evaluate
        rule_2: the second decision rule to evaluate
        data: the data used to run the test
    
    Returns:
        An estimate of the asymptotic covariance.
    """
    x = data[:,2:]
    y = data[:,0]
    d = data[:,1]
    #g = (y*d/prop - y*(1-d)/(1 - prop))
    g_1 = y*d/prop
    g_2 = y*(1-d)/(1-prop)
    is_in_1 = np.array([PWM._is_in(rule_1, i) for i in x])
    is_in_2 = np.array([PWM._is_in(rule_2, i) for i in x])
    part_1 = np.mean((g_1*g_1)*is_in_1*is_in_2 + (g_1*g_2)*is_in_1*(1 - is_in_2) + \
                     (g_1*g_2)*(1 - is_in_1)*is_in_2 + (g_2*g_2)*(1 - is_in_1)*(1 - is_in_2))
    part_2 = np.mean(g_1*is_in_1 + g_2*(1 - is_in_1))*np.mean(g_1*is_in_2 + g_2*(1 - is_in_2))
    return part_1 - part_2

def treat_mass(rule, data):
    """Computes an estimate for the mass of people being treated (used to
    define the random baseline below).
    """
    x = data[:,2:]
    return np.mean([PWM._is_in(rule,i) for i in x])
    
def random_baseline(mass, itera, data):
    """Computes the estimated welfare and standard error for an average of 
    random baseline rules as described in Appendix B.
    
    Args:
        mass: the mass of the sample to be treated
        itera: the number of random samples over which to average
        data: the data to be used to evaluate the rule
    """
    y = data[:,0]
    d = data[:,1]
    g_1 = y*d/prop
    g_2 = y*(1-d)/(1-prop)
    welfare = np.zeros(itera)
    welfare_matrix = np.zeros((itera, len(y)))
    for j in range(itera):
        random_baseline_list = np.random.binomial(1, mass, len(data))
        welfare_list = g_1*random_baseline_list + g_2*(1 - random_baseline_list)
        welfare[j] = np.mean(welfare_list)
        welfare_matrix[j,:] = welfare_list - welfare[j]
    cov_matrix = np.zeros((itera, itera))
    for i in range(len(y)):
        cov_matrix = cov_matrix + np.outer(welfare_matrix[:,i], welfare_matrix[:, i])
    return np.mean(welfare), \
1/(itera*np.sqrt(len(y)))*(np.sqrt(1/len(y)*np.dot(np.ones(itera), cov_matrix.dot(np.ones(itera)))))
""" END OF FUNCTION DEFINITIONS"""  

#run EWM   
#set level for EWM (class G_k)
level = 4
print("EWM")
g, z, b = PWM.EWM(data_est, prop, level, message = 1)
b_EWM = np.array(b)
b_EWM[1:] = b_EWM[1:]*10000

#run PWM
print("PWM1")
c1, z1, b1 = PWM.PWM(data_est, int(len(data_est)/4), prop, 1, 1)
print("PWM2")
c2, z2, b2 = PWM.PWM(data_est, int(len(data_est)/4), prop, 2, 1)
print("PWM3")
c3, z3, b3 = PWM.PWM(data_est, int(len(data_est)/4), prop, 3, 1)
print("PWM4")
c4, z4, b4 = PWM.PWM(data_est, int(len(data_est)/4), prop, 4, 1)

c = max(c1, c2, c3, c4)

if c == c1:
    b_PWM = np.array(b1)
    b_PWM[1:] = b_PWM[1:]*10000
elif c == c2:
    b_PWM = np.array(b2)
    b_PWM[1:] = b_PWM[1:]*10000
elif c == c3:
    b_PWM = np.array(b3)
    b_PWM[1:] = b_PWM[1:]*10000
else:
    b_PWM = np.array(b4)
    b_PWM[1:] = b_PWM[1:]*10000

    
#specify number of iterations for random baseline
itera = 1000
#compute random baseline
mass = treat_mass(b_PWM, data_test)
mass_first = treat_mass(b_PWM, data_first)
baseline_welfare, baseline_SE = random_baseline(mass, itera, data_test)
baseline_welfare_first, _ = random_baseline(mass_first, itera, data_first)

print("PWM welfare first half:", mean_welfare(b_PWM, data_first))
print("EWM welfare first half:", mean_welfare(b_EWM, data_first))
print("Baseline welfare first half:", baseline_welfare_first)
print("PWM welfare second half:", mean_welfare(b_PWM, data_test), " PWM SE:", \
      np.sqrt(var_est(b_PWM, data_test))/np.sqrt(len(data_test)))
print("EWM welfare second half:", mean_welfare(b_EWM, data_test), " EWM SE:",\
      np.sqrt(var_est(b_EWM, data_test))/np.sqrt(len(data_test)))
print("Baseline welfare second half:", baseline_welfare, " Baseline SE:",\
      baseline_SE)
print("Differece-in-Means p-value PWM vs EWM (one-tailed):", \
      1 - norm.cdf(welfare_difference(b_PWM, b_EWM, 0, data_test)))


    
