"""
This file contains all of the necessary functions to perform EMW and PWM for
threshold allocations, as described in Appendix C.

NOTE: Gurobi is required to run these functions.
"""

import numpy as np
from gurobipy import *

def gen_data(n, prop, reveal = 0):
    """Generates the data for the simulation exercise in the appendix.
    
    Args:
        n: sample size
        prop: the propensity score P(D=1)
        reveal: a boolean flag which specifies if (Y, D) should be returned
            (as would be the case in reality), or if (Y(1), Y(0)) should be returned.
        
    Returns:
        A numpy array which corresponds to the generated data.
    """
    x1 = np.random.uniform(0,1,n)
    x2 = np.random.uniform(0,1,n)
    x3 = np.random.uniform(0,1,n)
    x4 = np.random.uniform(0,1,n)
    x5 = np.random.uniform(0,1,n)
    y1 = 50*(2*x2 - (1-x1)**4 - 0.5 + 0.5*(x3 -x4)) + np.random.uniform(-80,80,n)
    y0 = 25*(x3-x4) +  np.random.uniform(-80,80,n)
    d = np.random.binomial(1,prop,n)
    y = (y1*d + y0*(1-d))
    if reveal == 0:
        return np.dstack((y,d,x1,x2,x3,x4,x5))[0]
    else:
        return np.dstack((y1,y0,x1,x2,x3,x4,x5))[0]
            
def EWM(data, prop, level = 1, rule = 1, message = 0, dm = 0):
    """Performs EWM over threshold allocations for a given number of covariates,
    as described in the appendix.
    
    Args:
        data: a numpy array corresponding to dataset to be used, in the format
        (Y, D, X)
        prop: the propensity score for the experiment P(D = 1)
        level: specifies the number of "active" covariates k
        rule: a boolean variable which separates the "zero threshold" case 
            (treat everyone/no one) from the "positive threshold" case.
        message: boolean flag which specifies if gurobi output should be printed
        dm: boolean flag which specifies if outcomes should be demeaned as in 
            Remark 3.2
        
    Returns:
        A vector which correponds to tau_i for each individual in the sample
        A vector which corresponds to the assignment of each individual in the sample
        A numpy array which corresponds to the optimal rule
    """
    n = len(data)
    x = data[:,2:]
    dim = len(x[0])
    thresh_list = list(range(0,dim,1))
    y = data[:,0]
    ym = np.mean(y)
    #this is the de-meaned version of the outcome
    ydm = y - ym
    d = data[:,1]
    if dm == 0:
        g = (1/n)*(y*d/prop - y*(1-d)/(1-prop))
    else:
        g = (1/n)*(ydm*d/prop - ydm*(1-d)/(1-prop))
    
    C = 12
    if message == 1:
        print("setup...")
        
    if rule == 0: 
        obj = np.sum(g)
        if obj > 0:
            if message == 1:
                print("all")
                print("objective:",np.dot(g,np.ones(n)))
            return g, np.ones(n), np.array([np.concatenate(([1], np.zeros(dim)))])
        else:
            if message == 1:
                print("nothing")
                print("objective:",np.dot(g,np.zeros(n)))
            return g, np.zeros(n), np.array([np.concatenate(([-1], np.zeros(dim)))])
        
    elif rule == 1:
        # create the LP object, set up as a maximization problem
        m = Model("EWM")
        if message != 1:
            m.setParam( 'OutputFlag', False )
            
        index_vector = list(range(0,n,1))
        beta_dim = list(range(0,dim,1))
    
        # set up decision variables
        z_star = {}
        for iv in index_vector:
            z_star[iv] = m.addVar(vtype = GRB.BINARY,name="z_star%s"%iv)  
        z = {}
        for iv in index_vector:
            for j in thresh_list:
                z[iv,j] = m.addVar(vtype = GRB.BINARY,name="z%s%s"%(iv,j))
        const = {}
        for j in range(0,dim,1):
            const[j] = m.addVar(ub=10,lb=-10,name="const%s"%j)
        beta = {}
        for d in beta_dim:
            for j in thresh_list:
                beta[d,j] = m.addVar(ub=1, lb=-1,name="beta%s%s"%(d,j))
        y1 = {}
        y2 = {}
        e = {}
        for j in thresh_list:
            y1[j] = m.addVar(vtype = GRB.BINARY,name="y1%s"%j)
            y2[j] = m.addVar(vtype = GRB.BINARY,name="y2%s"%j)
            e[j] = m.addVar(vtype = GRB.BINARY, name="e%s"%j)
        
        # set up tuning parameters for Gurobi
        m.setParam('MIPFocus',2)
        m.setParam('Cuts',0)
        m.setParam('Method',2)
        m.setParam('Heuristics',0)
        m.setParam('VarBranch', 0)
        m.setParam('IntFeasTol',1e-9)
        m.setParam('FeasibilityTol',1e-9)
            
        m.update()
        
        # set objective
        m.setObjective(quicksum(g[iv]*z_star[iv] for iv in index_vector),GRB.MAXIMIZE)
        
        # set up constraints
        for iv in index_vector:
            X = np.concatenate(([1],x[iv]))
            for j in thresh_list:
                B = np.concatenate(([const[j]],[beta[d,j] for d in beta_dim]))    
                m.addConstr(np.dot(X,B) - z[iv,j]*C + C*10**(-7) <= 0)
                m.addConstr(C*z[iv,j] - (C + np.dot(X,B)) <= 0)
            m.addConstr(1 - dim + quicksum(z[iv,j] for j in thresh_list) <= z_star[iv])
            m.addConstr(z_star[iv] <= (1/dim)*quicksum(z[iv,j] for j in thresh_list))
        for j in thresh_list:
            m.addConstr((-1)*e[j]<=const[j])
            m.addConstr(const[j]<=e[j])
            for d in beta_dim:
                if j == d:
                    m.addConstr(beta[j,j] == (-1)*y1[j] + y2[j])
                else:
                    m.addConstr(beta[d,j] == 0)
            m.addConstr(y1[j] + y2[j] == e[j]) 
        m.addConstr(quicksum(e[j] for j in thresh_list) == level)
        
        m.optimize()
 
        z_vec = [z_star[iv].x for iv in index_vector]
        b_vec = {}
        for j in thresh_list:
            b_vec[j] = [const[j].x] + [beta[i,j].x for i in beta_dim]
        return g, z_vec, _remove_zero_row(list(b_vec.values()))
    
def PWM(data, m, prop, message = 0, dm = 0):
    """
    Performs PWM over the sequence of threshold allocations for varying k
    
    Args:
        data: a numpy array corresponding to the dataset in (Y, D, X) form
        m: The size of the test dataset
        prop: The propensity score P(D = 1)
        message: boolean flag which specifies if the gurobi output should be
            printed
        dm: a boolean flag which specifies if the outcome should be demeaned as
            in Remark 3.2
        
    Returns:
        The index of the selected class (i.e. k)
        The resulting policy
    """
    def _pen(data, rule, prop):
        """Function which returns the value of the holdout penalty"""
        y = data[:,0]
        ym = np.mean(y)
        ydm = y - ym
        d = data[:,1]
        if dm == 0:
            g = (y*d/prop - y*(1-d)/(1-prop))
        else:
            g = (ydm*d/prop - ydm*(1-d)/(1-prop))
        z = np.zeros(len(data))
        dim = len(rule)
        for i in range(0,len(data),1):
            covs = np.concatenate(([1], data[i,2:]))
            if np.array_equal(np.dot(rule,covs) >= 0, np.ones(dim)):
                z[i] = 1
            else:
                z[i] = 0     
        return np.dot(g,z)
    
    n = len(data)
    len_train_data = n-m
    train_data = data[0:len_train_data]
    pen_data = data[len_train_data:n]
    
    [g_0,z_0,b_0] = EWM(train_data, prop, rule = 0)    
    [g_1,z_1,b_1] = EWM(train_data, prop, 1, 1)
    [g_2,z_2,b_2] = EWM(train_data, prop, 2, 1)
    [g_3,z_3,b_3] = EWM(train_data, prop, 3, 1)
    [g_4,z_4,b_4] = EWM(train_data, prop, 4, 1)
    [g_5,z_5,b_5] = EWM(train_data, prop, 5, 1)
        
    W_0 = _pen(pen_data,b_0,prop)
    W_1 = _pen(pen_data,b_1,prop)
    W_2 = _pen(pen_data,b_2,prop)
    W_3 = _pen(pen_data,b_3,prop)
    W_4 = _pen(pen_data,b_4,prop)
    W_5 = _pen(pen_data,b_5,prop)
    
    W = max(W_0, W_1, W_2, W_3, W_4, W_5)
    
    if W == W_0:
        return[0, b_0]
    elif W == W_1:
        return [1, b_1]
    elif W == W_2:
        return [2, b_2]
    elif W == W_3:
        return [3, b_3]
    elif W == W_4:
        return [4, b_4]
    else:
        return [5, b_5]
        
def evaluate_rule(rule, prop, acc = 50000):
    """Function used to evaluate the welfare of a given decision rule.
    Args: 
        rule: the rule to be evaluated
        prop: the propensity score P(D  = 1)
        acc: the auxiliary sample size to generate
    
    Returns:
        an estimate of the welfare of the given policy
    """
    expect = np.zeros(acc)
    #proportion doesn't matter here
    data = gen_data(acc, prop, reveal = 1)
    dim = len(rule)
    for i in range(0, acc, 1):
        covs = np.concatenate(([1], data[i,2:]))      
        if np.array_equal(np.dot(rule, covs) >= 0, np.ones(dim)):
            expect[i] = data[i,0]
        else:
            expect[i] = data[i,1]
    return np.mean(expect)  

def _remove_zero_row(rule):
    """auxiliary function to remove rows of zeros in an array"""
    rule = np.array(rule)
    return rule[~np.all(np.abs(rule) < 0.00001 , axis = 1)]


